package infovis.util;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

import javax.imageio.ImageIO;
import javax.swing.JComponent;

import org.apache.batik.dom.GenericDOMImplementation;
import org.apache.batik.svggen.CachedImageHandlerBase64Encoder;
import org.apache.batik.svggen.GenericImageHandler;
import org.apache.batik.svggen.SVGGeneratorContext;
import org.apache.batik.svggen.SVGGraphics2D;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;

/**
 * Saves screenshots.
 * 
 * @author Joschi <josua.krause@googlemail.com>
 */
public final class Screenshot {

  /** The scaling factor for raster screenshots. */
  private static final int SCALE = 3;

  /** No constructor. */
  private Screenshot() {
    // no constructor
  }

  /**
   * Ensures a directory.
   * 
   * @param dir The directory.
   */
  private static void ensureDir(final File dir) {
    if(dir.isFile()) return;
    if(dir.exists()) return;
    ensureDir(dir.getParentFile());
    dir.mkdir();
  }

  /**
   * Creates an output stream for an image.
   * 
   * @param dir The directory.
   * @param prefix The name prefix.
   * @param postfix The name postfix.
   * @return The output.
   * @throws IOException I/O Exception.
   */
  private static OutputStream createOutput(final File dir, final String prefix,
      final String postfix) throws IOException {
    ensureDir(dir);
    final File out;
    int i = 0;
    for(;;) {
      final File f = new File(dir, prefix + (i++) + postfix);
      if(!f.exists()) {
        out = f;
        break;
      }
    }
    return new FileOutputStream(out);
  }

  /**
   * Saves a screenshot as SVG.
   * 
   * @param dir The directory.
   * @param prefix The name prefix.
   * @param comp The component to draw.
   * @throws IOException I/O Exception.
   */
  public static void saveSVG(final File dir, final String prefix, final JComponent comp)
      throws IOException {
    saveSVG(createOutput(dir, prefix, ".svg"), comp);
  }

  /**
   * Saves a screenshot as PNG.
   * 
   * @param dir The directory.
   * @param prefix The name prefix.
   * @param comp The component to draw.
   * @throws IOException I/O Exception.
   */
  public static void savePNG(final File dir, final String prefix, final JComponent comp)
      throws IOException {
    savePNG(createOutput(dir, prefix, ".png"), comp);
  }

  /**
   * Saves a screenshot as SVG.
   * 
   * @param out The output.
   * @param comp The component.
   * @throws IOException I/O Exception.
   */
  public static void saveSVG(final OutputStream out, final JComponent comp)
      throws IOException {
    final DOMImplementation domImpl = GenericDOMImplementation.getDOMImplementation();
    final String svgNS = "http://www.w3.org/2000/svg";
    final Document document = domImpl.createDocument(svgNS, "svg", null);
    final SVGGeneratorContext ctx = SVGGeneratorContext.createDefault(document);
    ctx.setEmbeddedFontsOn(true);
    ctx.setComment("generated by BusVis");
    // Reuse our embedded base64-encoded image data.
    final GenericImageHandler ihandler = new CachedImageHandlerBase64Encoder();
    ctx.setGenericImageHandler(ihandler);
    final SVGGraphics2D g = new SVGGraphics2D(ctx, true);
    final Rectangle rect = comp.getVisibleRect();
    g.setSVGCanvasSize(new Dimension(rect.width, rect.height));
    comp.paintAll(g);
    g.dispose();
    final Writer writer = new OutputStreamWriter(out, "UTF-8");
    g.stream(writer, true);
    writer.close();
  }

  /**
   * Saves a screenshot as PNG.
   * 
   * @param out The output.
   * @param comp The component.
   * @throws IOException I/O Exception.
   */
  public static void savePNG(final OutputStream out, final JComponent comp)
      throws IOException {
    final Rectangle rect = comp.getVisibleRect();
    final BufferedImage img = new BufferedImage(
        rect.width * SCALE, rect.height * SCALE, BufferedImage.TYPE_INT_ARGB);
    final Graphics2D g = img.createGraphics();
    g.scale(SCALE, SCALE);
    comp.paintAll(g);
    g.dispose();
    ImageIO.write(img, "png", out);
    out.close();
  }

}
